package in.playsimple.common.flutter;

import android.app.Activity;
import android.os.CountDownTimer;
import android.util.Log;

import org.json.JSONObject;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import in.playsimple.Constants;
import in.playsimple.GameSpecific;
import in.playsimple.PSPNGameSpecific;
import in.playsimple.common.Analytics;
import in.playsimple.common.Experiment;
import in.playsimple.common.PSPrivacy;
import in.playsimple.common.PSUtil;
import in.playsimple.common.Track;
import in.playsimple.common.User;
import in.playsimple.common.mopub.PSMopubAds;
import in.playsimple.pspn.XPromoUtil;
import io.flutter.embedding.engine.FlutterEngine;
import io.flutter.embedding.engine.FlutterEngineCache;
import io.flutter.embedding.engine.dart.DartExecutor;
import io.flutter.plugin.common.BasicMessageChannel;
import io.flutter.plugin.common.BinaryMessenger;
import io.flutter.plugin.common.MethodChannel;
import io.flutter.plugin.common.StringCodec;

public class FlutterBridge {
    public static final String TAG = "FlutterBridge";
    public static Activity activity = null;
    static final String FLUTTER_ENGINE_ID = "ps_flutter_engine";

    public static FlutterEngine flutterEngine = null;

    private static MethodChannel _methodChannel;
    private static BasicMessageChannel<String> _messageChannel;
    private static Boolean _channelsInitialized = false;

    private static final String FLUTTER_METHOD_CHANNEL = "in.playsimple.flutter.method";
    private static final String FLUTTER_MESSAGE_CHANNEL = "in.playsimple.flutter.message";

    private static ArrayList<FlutterCriticalQueue> criticalQueue = null;

    public static void setActivity(Activity act) {
        Log.i(FlutterBridge.TAG, "flutter - engine is initialized from Java");
        activity = act;

        flutterEngine = new FlutterEngine(activity);

        // Pre-warm your FlutterEngine by starting Dart execution.
        flutterEngine
                .getDartExecutor()
                .executeDartEntrypoint(
                        DartExecutor.DartEntrypoint.createDefault()
                );

        FlutterEngineCache
                .getInstance()
                .put(FlutterBridge.FLUTTER_ENGINE_ID, flutterEngine);

        initializeChannels();
    }

    public static void sendDataToFlutterModule(String params, boolean isActivity) {
        if (!_channelsInitialized) {
            Log.i(FlutterBridge.TAG, "Channels not yet initialized. Message not sent. Message=" + params);
            return;
        }

        if (isActivity) {
            FlutterViewActivity.startActivity(activity, params);
        } else {
            activity.runOnUiThread(() -> _messageChannel.send(params));
        }
    }

    public static void sendDataToFlutterModule(String params, boolean isActivity, boolean isCritical) {
        if (isCritical) {
            if (!_channelsInitialized) {
                addToCriticalQueue(params, isActivity, false);
                return;
            }
        }

        sendDataToFlutterModule(params, isActivity);
    }

    public static void sendStateToFlutterModule(String controller, String action) {
        JSONObject msgInfo = new JSONObject();
        try {
            msgInfo.put("route", controller);
            msgInfo.put("action", action);
        } catch (Exception e) {

        }
        sendDataToFlutterModule(msgInfo.toString(), false);
    }

    private static void processCriticalQueue() {
        // Processing the critical queue with a delay between every message
        // We noticed in testing that certain messages are lost over the channel when sent synchronously
        if (criticalQueue != null && criticalQueue.size() > 0) {
            long totalTime = 1000 * (criticalQueue.size() + 5); // Additional 5 ticks to ensure it finishes incase onTick takes more time
            Iterator<FlutterCriticalQueue> iter = criticalQueue.iterator();

            new CountDownTimer(totalTime, 1000) {

                private boolean firstTick = false;

                public void onTick(long millisUntilFinished) {
                    if (firstTick) { // Skipping the first tick as it happens immediately after start time
                        action();
                    }
                    firstTick = true;
                }

                public void onFinish() {
                    action();
                }

                private void action() {
                    if (iter.hasNext()) {
                        FlutterCriticalQueue obj = iter.next();
                        Log.i(FlutterBridge.TAG, "Processing Critical Queue params=" + obj.params);
                        sendDataToFlutterModule(obj.params, obj.isActivity);
                        iter.remove();
                    }
                }
            }.start();
        }
    }

    private static Boolean initializeChannels() {
        if (flutterEngine == null) {
            Log.e(FlutterBridge.TAG, "Channels being initialized before FlutterEngine, ensure FlutterBridge.setActivity is called first");
            return false;
        }
        Log.i(FlutterBridge.TAG, "Attempting to initialize flutter channels");

        if (_channelsInitialized) {
            return true;
        }

        // https://github.com/flutter/flutter/issues/41102#issuecomment-557738290
        // Use the GeneratedPluginRegistrant to add every plugin that's in the pubspec.
        // GeneratedPluginRegistrant.registerWith(flutterEngine);

        BinaryMessenger messenger = FlutterEngineCache.getInstance().get(FlutterBridge.FLUTTER_ENGINE_ID).getDartExecutor();
        _methodChannel = new MethodChannel(messenger, FLUTTER_METHOD_CHANNEL);
        _messageChannel = new BasicMessageChannel<>(messenger, FLUTTER_MESSAGE_CHANNEL, StringCodec.INSTANCE);

        _methodChannel.setMethodCallHandler(
                (call, result) -> {
                    // Note: this method is invoked on the main thread.
                    String method = call.method;
                    Log.i(FlutterBridge.TAG, "On method call triggered:" + method);
                    switch (method) {
                        case "initializeCustomView":
                            FlutterViewActivity.startActivity(activity, "");
                            result.success(null);
                            break;
                        case "showNativeToast":
                            String msg = call.argument("msg");
                            PSUtil.showMessage(msg);
                            result.success(null);
                            break;
                        case "getBatteryLevel":
                            int batteryLevel = PSUtil.getBatteryLevel();

                            if (batteryLevel != -1) {
                                result.success(batteryLevel);
                            } else {
                                result.error("UNAVAILABLE", "Battery level not available.", null);
                            }
                            break;
                        case "updateNameEmail":
                            final String name = call.argument("name");
                            final String email = call.argument("email");
                            try {
                                User user = User.get();
                                user.updateNameEmail(name, email);
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        case "updateChosenCountry":
                            final String chosenCountry = call.argument("cc");
                            try {
                                User user = User.get();
                                user.updateChosenCountry(chosenCountry);
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        case "updateAgeRange":
                            final String cmin = call.argument("cmin");
                            final String cmax = call.argument("cmax");
                            try {
                                User user = User.get();
                                user.updateAgeRange(cmin, cmax);
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        case "acceptPrivacyPolicy":
                            try {
                                User user = User.get();
                                user.acceptPrivacyPolicy();
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        case "sendTrackingFlutter":
                            final String k = call.argument("k");
                            final String p = call.argument("p");
                            final String c = call.argument("c");
                            final String o = call.argument("o");
                            final String f = call.argument("f");
                            final String g = call.argument("g");
                            String sp = call.argument("s");
                            String co = call.argument("co");
                            String app = call.argument("app");
                            try {
                                if (sp == null) {
                                    sp = "";
                                }

                                if (co == null || co.equals("")) {
                                    co = "1";
                                }

                                if (app == null) {
                                    app = "";
                                }

                                Track.trackCounter(k, p, c, o, f, g, sp, co, app);
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        case "mopub":
                            boolean s = PSMopubAds.handleDartCall(call);
                            Log.i(FlutterBridge.TAG, "Handling dart call for mopub:" + s);
                            result.success(s);
                            break;
                        case "logFlutterException":
                            final String expMessage = call.argument("e");
                            final String expStack = call.argument("s");
                            Analytics.sendReport(expStack);

                            Exception exp = new Exception(expMessage);
                            try {
                                // Custom parsing logic to convert the stack string into meaningful information at top of the stack
                                if (expStack.length() > 0) {
                                    String[] parts = expStack.split("\n")[0].replace("#0", "").replace(" ", "").split("[\\(\\)\\:]+");
                                    if (parts.length >= 4) {
                                        String[] subparts = parts[0].split("\\.");
                                        if (subparts.length >= 2) {
                                            StackTraceElement[] trace = new StackTraceElement[]{
                                                    new StackTraceElement(subparts[0],
                                                            subparts[1],
                                                            parts[2],
                                                            Integer.parseInt(parts[3]))
                                            };
                                            exp.setStackTrace(trace);
                                        }
                                    }
                                }
                            } catch (Exception e) {
                            }
                            Analytics.logException(exp);
                            result.success(1);
                            break;
                        case "experiment":
                            String variant = Experiment.getChosenVariant(call.argument("name"));
                            Log.i(FlutterBridge.TAG, "Handling dart call for experiment variant:" + variant);
                            result.success(variant);
                            break;
                        case "data_privacy":
                            try {
                                PSPrivacy.sendDataPrivacyStatesToGame(call.arguments.toString());
                                result.success(1);
                            } catch (Exception e) {
                                result.error("Exception", e.getMessage(), null);
                            }
                            break;
                        default:
                            if (!PSPNFlutterBridge.handleDartCall(call, result)) {
                                result.notImplemented();
                            }
                            break;
                    }
                });

        _channelsInitialized = true;

        JSONObject config = new JSONObject();
        try {
            config.put("route", "config");
            config.put("g", Constants.GAME_ID + "");
            config.put("c", Constants.CLIENT + "");
        } catch (Exception e) {
        }
        addToCriticalQueue(config.toString(), false, true);

        FlutterBridge.processCriticalQueue();

        return true;
    }

    private static void addToCriticalQueue(String params, boolean isActivity, boolean atStart) {
        FlutterCriticalQueue obj = new FlutterCriticalQueue(params, isActivity);
        if (criticalQueue == null) {
            criticalQueue = new ArrayList<>();
        }

        if (atStart) {
            criticalQueue.add(0, obj);
        } else {
            criticalQueue.add(obj);
        }
    }

    public static class FlutterCriticalQueue {

        public String params;

        public boolean isActivity;

        FlutterCriticalQueue(String params, Boolean isActivity) {
            this.params = params;
            this.isActivity = isActivity;
        }
    }
}