/**
 * quest flags are used to change required level for daily quests when any event of experiment is enabled. We will be passing an integer for each bucket for which we want to enable changes, this integer will then be used as a multiplier for required level
 * @constructor
 */
QuestFlags = function () {
    var EVENT_ON = 1;
    var EVENT_OFF = 0;

    var questEventsFlagsMap = {}; // saving in quest data, it stores what all events are currently active
    var multiplierForBucket = {}; // saving in quest data, it stores the multiplier for each bucket.

    this.init = function () {

    };

    /**
     * loading questEventsFlagsMap and multiplierForBucket
     * @param quests quest data map, which contains all the data of quest variables
     */
    this.load = function (quests) {
        if (quests.questEventsFlagsMap != undefined) {
            questEventsFlagsMap = quests.questEventsFlagsMap;
        }
        if (quests.multiplierForBucket != undefined) {
            multiplierForBucket = quests.multiplierForBucket;
        }
    };

    /**
     * saving questEventsFlagsMap and multiplierForBucket
     * @param quests saving questEventsFlagsMap and multiplierForBucket in quest file data
     * @returns {map} quest data
     */
    this.save = function (quests) {
        quests.questEventsFlagsMap = questEventsFlagsMap;
        quests.multiplierForBucket = multiplierForBucket;
        return quests;
    };

    /**
     * this function is called daily to get the multipliers for each daily quest bucket. it checks whether the experiment is active or not, event within the experiment is active or not.
     * There is priority associated with each event, multipliers with the lowest priority number is used.
     */
    this.getMultiplierForDailyQuests = function () {
        // since this function is called daily, we should first reset the multiplierForBucket map.
        questFlagsObj.resetMultiplierForBucket();
        
        // fetching required level multipliers which are runtime.
        var dailyQuestReqdLevelMultiplierRuntime = xpromoQuestRuntimesObj.getRuntime(xpromoConstantsObj.RUNTIME_DAILY_QUEST_REQUIRED_LEVEL);
        if (dailyQuestReqdLevelMultiplierRuntime == undefined) {
            return;
        }

        // initializing currentPriority as max value
        var currentPriority = Number.MAX_VALUE;

        // iterating through each event in dailyQuestReqdLevelMultiplierRuntime which has expName as key and event, priority, bucket Numbers and multipliers as value.
        for (var expName in dailyQuestReqdLevelMultiplierRuntime) {

            // getting the value as runtimeMapValue from dailyQuestReqdLevelMultiplierRuntime using expName as key
            var runtimeMapValue = dailyQuestReqdLevelMultiplierRuntime[expName];

            // the length of runtimeMapValue should always be greater than 2, as it contains event, priority, bucket Numbers and multiplier
            if (runtimeMapValue.length > 2) {

                // getting eventName and priority which are at index 0 and 1 respectively.
                var eventName = runtimeMapValue[0], priority = runtimeMapValue[1];

                // if priotiy is less than currentPriority only then populate required Level multiplers for each bucket. other checks are checking if event is avtive and exp is active
                if (priority < currentPriority && questFlagsObj.getEventFlag(eventName) == 1 && epmObj.getChosenVariant(expName) != constantsObj.EXP_VAR_CONTROL) {

                    // incrementing by 2, as bucketNumbers are at even indexes and required level multiplers are at odd indexes
                    for (var i = 2; i < runtimeMapValue.length; i = i + 2) {
                        // runtimeMapValue[i] contains bucket number and runtimeMapValue[i + 1] contains multiplier, storing these values in multiplierForBucket for future use.
                        multiplierForBucket[runtimeMapValue[i]] = runtimeMapValue[i + 1];
                    }

                    // setting currentPriority as priority of event that was executed, so that only lower priority events are executed at each iteration
                    currentPriority = priority;
                }
            }
        }
    };

    /**
     * it is used to fetch daily quest required level multiplier. this multiplier is then multiplied with required level to change the it when any event is active.
     * @param bucketId bucket is an integer. multiplier for all the quests in a bucket will remain same.
     * @returns {double} returns a double value as a multiplier. if multiplierForBucket doesn't have multiplier for a bucket, it would simply return 1 as multiplier.
     */
    this.getMultiplierForBucket = function (bucketId) {
        if (multiplierForBucket[bucketId] == undefined) {
            return 1;
        }
        return multiplierForBucket[bucketId];
    };

    /**
     * set the event flag, if event is active.
     * @param eventName name of event which should be set as active.
     */
    this.setEventFlag = function (eventName) {
        questEventsFlagsMap[eventName] = EVENT_ON;
    };

    /**
     * reset the event flag, if event is no longer active.
     * @param eventName eventName name of event which is no longer active.
     */
    this.resetEventFlag = function (eventName) {
        questEventsFlagsMap[eventName] = EVENT_OFF;
    };

    /**
     * getting the value of event flag, if event is not defined it will return as off.
     * @param eventName name of the event
     * @returns {int} returns 0/1 whether an event is off or on
     */
    this.getEventFlag = function (eventName) {
        if(questEventsFlagsMap[eventName] == undefined) {
            return EVENT_OFF;
        }
        return questEventsFlagsMap[eventName];
    };

    /**
     * resets multiplierForBucket map. this function will be called daily once.
     */
    this.resetMultiplierForBucket = function () {
        multiplierForBucket = {};
    };
};